#include <algorithm>
#include <array>
#include <cstdint>
#include <random>
#include <thread>
#include <vector>
#include <chrono>
#include <iostream>
#include <list>
#include <mutex>
#include <thread>
#include <queue>
#include <atomic>
#include <condition_variable>
#include <iterator>

template <typename TimePoint>
std::chrono::milliseconds to_ms(TimePoint tp) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(tp);
}


using namespace std::chrono_literals;

volatile int dump = 0;

class waiter {
public:
    explicit waiter(int waitfactor): m_waitfactor{ waitfactor } {}

    waiter(const waiter& rhs): m_waitfactor(rhs.m_waitfactor) {
        slowdown();
    }

    waiter& operator=(const waiter& rhs) {
        m_waitfactor = rhs.m_waitfactor;
        slowdown();
        return *this;
    }

    int process() const {
        slowdown();
        return m_waitfactor;
    }

private:
    void slowdown() const {
        for (int i = 0; i < m_waitfactor; ++i) {
            auto func = [=](int d) {
                return i / (d + 1);
            };
            dump = func(dump);
        }
    }
    int m_waitfactor;
};

template <typename Numeric>
double get_median(std::vector<Numeric> data) {
    std::sort(begin(data), end(data));
    size_t mid = data.size() / 2;
    if (data.size() % 2 == 0) {
        // average of middle two
        return (data[mid] + data[mid - 1]) / 2.0;
    } else {
        return static_cast<double>(data[mid]);
    }
}

double generate_double() {
    static std::default_random_engine rng{ std::random_device{}() };
    static std::uniform_real_distribution<> dist(0, 1);
    return dist(rng);
}

std::vector<waiter> generate_unbalanced_data(int d_size, int splits, int workfactor, int factor) {
    std::vector<waiter> ret;
    ret.reserve(d_size);
    auto split_size = d_size / splits;
    for (int i = 0; i < d_size - split_size; ++i) {
        ret.emplace_back(workfactor);
    }
    for (int i = 0; i < split_size; ++i) {
        ret.emplace_back(workfactor * factor);
    }
    return ret;
}

std::vector<waiter> generate_even_data(int d_size, int splits, int workfactor, int factor) {
    std::vector<waiter> ret;
    ret.reserve(d_size);
    auto split_size = d_size / splits;
    auto num_harder = split_size / splits;
    for (int s = 0; s < splits; ++s) {
        for (int i = 0; i < split_size - num_harder; ++i) {
            ret.emplace_back(workfactor);
        }
        for (int i = 0; i < num_harder; ++i) {
            ret.emplace_back(workfactor * factor);
        }
    }
    return ret;
}

static const int num_iterations = 5;

long long process_MT_static(const std::vector<waiter>& data, int num_threads) {
    auto t1 = std::chrono::high_resolution_clock::now();
    std::vector<std::thread> threads; threads.reserve(num_threads);
    auto split_size = data.size() / num_threads;
    for (int i = 0; i < num_threads; ++i) {
        threads.emplace_back([&](int from, int to) {
            for (; from < to; ++from) {
                data[from].process();
            }
        }, i * split_size, (i + 1) * split_size);
    }

    for (auto& t : threads) {
        t.join();
    }

    auto t2 = std::chrono::high_resolution_clock::now();
    return to_ms(t2 - t1).count();
}

long long process_MT_dynamic(const std::vector<waiter>& data, int num_threads) {
    auto t1 = std::chrono::high_resolution_clock::now();
    std::vector<std::thread> threads; threads.reserve(num_threads);
    std::atomic<size_t> current_index{ 0 };
    for (int i = 0; i < num_threads; ++i) {
        threads.emplace_back([&](size_t limit) {
            size_t taken = 0;
            while (taken < limit) {
                taken = current_index++;
                if (taken < limit) {
                    data[taken].process();
                }
            }
        }, data.size());
    }

    for (auto& t : threads) {
        t.join();
    }

    auto t2 = std::chrono::high_resolution_clock::now();
    return to_ms(t2 - t1).count();
}

class index_drawer {
public:
    index_drawer():last{ 0u }, current{ 0u } {}
    index_drawer(uint32_t first, uint32_t last):last{ last }, current{ first } {}

    bool get(uint32_t& index) {
        if (current >= last) {
            return false;
        }
        uint32_t temp = current.fetch_add(1);
        if (temp >= last) {
            return false; //dclp
        }
        index = temp;
        return true;
    }

    void reset(uint32_t from, uint32_t to) {
        current = from;
        last = to;
    }

private:
    std::atomic<uint32_t> current;
    // Expected cache line size is 64B, we already use 4B of that.
    std::array<uint8_t, 60> frequent_access_padding;
    uint32_t last;
    // See above.
    std::array<uint8_t, 60> false_sharing_padding;
};

class index_cupboard {
public:
    index_cupboard(std::size_t number_of_drawers, uint32_t up_to):number_of_drawers{ number_of_drawers },
                                                                  drawers{ std::make_unique<index_drawer[]>(number_of_drawers) } {
        int part_size = up_to / number_of_drawers;
        int mod = up_to % number_of_drawers;
        for (int i = 0; i < number_of_drawers; ++i) {
            drawers[i].reset(i * part_size + std::min(i, mod),
                            (i + 1) * part_size + std::min(i + 1, mod));
        }
    }

    bool get(int starting_drawer, uint32_t& index, int& last_drawer) {
        uint32_t drawer = starting_drawer;
        for (int i = 0; i < number_of_drawers; ++i) {
            if (drawers[drawer].get(index)) {
                last_drawer = drawer;
                return true;
            }
            if (++drawer >= number_of_drawers) {
                drawer = 0;
            }
        }
        return false;
    }

private:
    std::unique_ptr<index_drawer[]> drawers;
    std::size_t number_of_drawers = 0;
};


long long process_MT_smart(const std::vector<waiter>& data, int num_threads) {
    auto t1 = std::chrono::high_resolution_clock::now();

    index_cupboard cupboard(num_threads, data.size());

    std::vector<std::thread> threads; threads.reserve(num_threads);
    for (int i = 0; i < num_threads; ++i) {
        threads.emplace_back([&](int cupboard_idx) {
            uint32_t data_idx = 0;
            while (cupboard.get(cupboard_idx, data_idx, cupboard_idx)) {
                data[data_idx].process();
            }
        }, i);
    }

    for (auto& t : threads) {
        t.join();
    }

    auto t2 = std::chrono::high_resolution_clock::now();
    return to_ms(t2 - t1).count();
}


int main(int argc, char** argv) {
    int num_size;
    if (argc == 2) {
        num_size = std::atoi(argv[1]);
    } else {
        num_size = 100'800'00;
    }

    int thread_num = std::thread::hardware_concurrency();

    std::cout << "Size,NumThreads,WorkFactor,Multiplier,StaticMTMedianBalancedMs,StaticMTMedianUnbalancedMs,DynamicMTMedianBalancedMs"
        << ",DynamicMTMedianUnbalancedMs,SmartMTMedianBalancedMs,SmartMedianUnbalancedMs\n";
    for (int workfactor : {10, 25, 50, 75, 100}) {
        for (int factor : {2, 4, 6, 8, 10, 12, 14, 16, 18, 20}) {
            auto even_data = generate_even_data(num_size, thread_num, workfactor, factor);
            auto unbalanced_data = generate_unbalanced_data(num_size, thread_num, workfactor, factor);
            std::vector<long long> times;
            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(process_MT_static(even_data, thread_num));
            }
            auto static_MT_even_median = get_median(times);
            times.clear();

            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(process_MT_static(unbalanced_data, thread_num));
            }
            auto static_MT_unbalanced_median = get_median(times);
            times.clear();

            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(process_MT_dynamic(even_data, thread_num));
            }
            auto dynamic_MT_even_median = get_median(times);
            times.clear();

            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(process_MT_dynamic(unbalanced_data, thread_num));
            }
            auto dynamic_MT_unbalanced_median = get_median(times);
            times.clear();


            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(process_MT_smart(even_data, thread_num));
            }
            auto smart_MT_even_median = get_median(times);
            times.clear();

            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(process_MT_smart(unbalanced_data, thread_num));
            }
            auto smart_MT_unbalanced_median = get_median(times);
            times.clear();


            // Get median
            std::cout << num_size << ',' << thread_num << ',' << workfactor << ',' << factor
                << ',' << static_MT_even_median << ',' << static_MT_unbalanced_median << ',' << dynamic_MT_even_median
                << ',' << dynamic_MT_unbalanced_median << ',' << smart_MT_even_median << ',' << smart_MT_unbalanced_median << std::endl;
        }
   }
}
