#include <iostream>
#include <thread>
#include <array>
#include <vector>
#include <string>
#include <atomic>
#include <mutex>
#include <chrono>
#include <random>
#include <algorithm>
#include <numeric>
#include <iterator>



template <typename TimePoint>
std::chrono::milliseconds to_ms(TimePoint tp) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(tp);
}


double generate_double() {
    static std::default_random_engine rng{ std::random_device{}() };
    static std::uniform_real_distribution<> dist(0, 1);
    return dist(rng);
}

std::vector<double> generate_data(int d_size) {
    std::vector<double> ret;
    ret.reserve(d_size);
    std::generate_n(std::back_inserter(ret), d_size, generate_double);
    return ret;
}

static const int num_iterations = 5;

template <typename Numeric>
double get_median(std::vector<Numeric> data) {
    std::sort(begin(data), end(data));
    size_t mid = data.size() / 2;
    if (data.size() % 2 == 0) {
        // average of middle two
        return (data[mid] + data[mid-1])/ 2.0;
    } else {
        return static_cast<double>(data[mid]);
    }
}


struct packed_result {
    packed_result(int res = 0):res{ res } {}
    int res;
};

struct padded_result {
    padded_result(int res = 0):res{ res } {}
    int res;
    std::array<char, 60> padding;
};


bool validate(const std::vector<packed_result>& baseline, const std::vector<packed_result>& r1,
              const std::vector<packed_result>& r2, const std::vector<padded_result>& r3) {
    auto cmp = [](const auto& lhs, const auto& rhs) {
        return lhs.res == rhs.res;
    };
    return std::equal(begin(baseline), end(baseline), begin(r1), end(r1), cmp) &&
        std::equal(begin(baseline), end(baseline), begin(r2), end(r2), cmp) &&
        std::equal(begin(baseline), end(baseline), begin(r3), end(r3), cmp);
}

int main(int argc, char** argv) {
    int num_size;
    if (argc == 2) {
        num_size = std::atoi(argv[1]);
    } else {
        num_size = 100'800'00;
    }
    auto data = generate_data(num_size);
    std::cout << "ProblemSize,Threshold,NumThreads,STMedianMs,MTPackedImmediateMedianMs,MTPackedLocalMedianMs,MTPaddedImmediateMedianMs\n";
    for (double threshold : {1.0, 1.5, 2.0, 2.5, 3.0}) {

        // LCM: 10080
        for (int num_threads : {2, 4, 8, 12, 16, 20, 24, 28, 32, 36}) {
            size_t num_splits = num_threads;
            size_t split_size = num_size / num_threads;

            std::vector<long long> times; times.reserve(num_iterations);
            std::vector<packed_result> proper_results(num_splits);

            for (int iter = 0; iter < num_iterations; ++iter) {
                auto split_size = num_size / num_splits;
                auto t1 = std::chrono::high_resolution_clock::now();
                auto start = begin(data);
                for (int split = 0; split < num_splits; ++split) {
                    proper_results[split].res = std::count_if(start, start + split_size, [=](double d) {
                        if (exp(d) > threshold) {
                            return true;
                        }
                        return false;
                    });
                    start += split_size;
                }
                auto t2 = std::chrono::high_resolution_clock::now();
                times.push_back(to_ms(t2 - t1).count());
            }

            auto ST_median = get_median(std::move(times));
            times.clear();

            std::vector<packed_result> dumb_result(num_splits);
            std::vector<packed_result> dumb_result2(num_splits);
            std::vector<padded_result> smart_result(num_splits);

            auto thread_func1 = [&, threshold](int order, size_t from, size_t to) {
                for (; from < to; ++from) {
                    if (exp(data[from]) > threshold) {
                        dumb_result[order].res += 1;
                    }
                }
            };

            auto thread_func2 = [&, threshold](int order, size_t from, size_t to) {
                int partial_result = 0;
                for (; from < to; ++from) {
                    if (exp(data[from]) > threshold) {
                        partial_result += 1;
                    }
                }
                dumb_result2[order].res = partial_result;
            };

            auto thread_func3 = [&, threshold](int order, size_t from, size_t to) {
                for (; from < to; ++from) {
                    if (exp(data[from]) > threshold) {
                        smart_result[order].res += 1;
                    }
                }
            };

            std::vector<std::thread> threads;
            threads.reserve(num_threads);

            for (int iter = 0; iter < num_iterations; ++iter) {
                dumb_result.clear(); dumb_result.resize(num_splits);
                auto t1 = std::chrono::high_resolution_clock::now();
                for (int i = 0; i < num_threads; ++i) {
                    threads.emplace_back(thread_func1, i, i * split_size, (i + 1) * split_size);
                }
                for (auto& t : threads) {
                    t.join();
                }
                auto t2 = std::chrono::high_resolution_clock::now();
                times.push_back(to_ms(t2 - t1).count());
                threads.clear();
            }

            auto packed_imm_median = get_median(std::move(times));
            times.clear();

            for (int iter = 0; iter < num_iterations; ++iter) {
                dumb_result2.clear(); dumb_result2.resize(num_splits);
                auto t1 = std::chrono::high_resolution_clock::now();
                for (int i = 0; i < num_threads; ++i) {
                    threads.emplace_back(thread_func2, i, i * split_size, (i + 1) * split_size);
                }
                for (auto& t : threads) {
                    t.join();
                }
                auto t2 = std::chrono::high_resolution_clock::now();
                times.push_back(to_ms(t2 - t1).count());
                threads.clear();
            }
            auto packed_local_median = get_median(std::move(times));
            times.clear();

            for (int iter = 0; iter < num_iterations; ++iter) {
                smart_result.clear(); smart_result.resize(num_splits);
                auto t1 = std::chrono::high_resolution_clock::now();
                for (int i = 0; i < num_threads; ++i) {
                    threads.emplace_back(thread_func3, i, i * split_size, (i + 1) * split_size);
                }
                for (auto& t : threads) {
                    t.join();
                }
                auto t2 = std::chrono::high_resolution_clock::now();
                times.push_back(to_ms(t2 - t1).count());
                threads.clear();
            }
            auto padded_imm_median = get_median(std::move(times));
            times.clear();

            if (validate(proper_results, dumb_result, dumb_result2, smart_result)) {
                std::cout << num_size << ',' << threshold << ',' << num_threads << ','
                    << ST_median << ',' << packed_imm_median << ',' << packed_local_median << ','
                    << padded_imm_median << std::endl;
            } else {
                std::cerr << "invalid results!\n";
                for (int i = 0; i < num_splits; ++i) {
                    std::cerr << proper_results[i].res << ' ' << dumb_result[i].res << ' ' << dumb_result2[i].res << ' ' << smart_result[i].res << '\n';
                }
            }
        }
    }
}
