#include <algorithm>
#include <thread>
#include <vector>
#include <chrono>
#include <iostream>
#include <list>
#include <mutex>
#include <thread>
#include <queue>
#include <atomic>
#include <condition_variable>

template <typename TimePoint>
std::chrono::milliseconds to_ms(TimePoint tp) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(tp);
}

template <typename T>
class naive_queue {
public:

    void push(const T& elem) {
        std::unique_lock<std::mutex> lock(m_mutex);

        m_data.push(elem);
        if (m_size++ == 0) {
            m_pop_var.notify_all();
        }
    }

    void pop(T& elem) {
        std::unique_lock<std::mutex> lock(m_mutex);

        m_pop_var.wait(lock, [&]() {
            return m_size > 0;
        });
        elem = std::move(m_data.front());
        m_data.pop();

        --m_size;
    }


private:
    size_t m_size = 0;
    std::queue<T> m_data;
    std::mutex m_mutex;
    std::condition_variable m_pop_var;
};

template <typename T>
class list_queue {
public:

    void push(const T& elem) {
        std::list<T> temp_data;
        temp_data.push_back(elem);

        // Enter critical section after we are ready to push the data quickly
        {
            std::unique_lock<std::mutex> lock(m_mutex);

            m_data.splice(m_data.end(), temp_data);
            if (m_size++ == 0) {
                m_pop_var.notify_all();
            }
        }
    }

    void pop(T& elem) {
        std::list<T> temp_data;

        // Enter critical data to slice an element
        {
            std::unique_lock<std::mutex> lock(m_mutex);

            m_pop_var.wait(lock, [&]() {
                return m_size > 0;
            });

            temp_data.splice(temp_data.end(), m_data, m_data.begin());
            --m_size;
        }

        elem = std::move(temp_data.front());
    }

private:
    std::size_t m_size = 0;
    std::list<T> m_data;
    std::mutex m_mutex;
    std::condition_variable m_pop_var;
};

using namespace std::chrono_literals;

volatile int dump = 0;

class waiter {
public:
    explicit waiter(int waitfactor): m_waitfactor{ waitfactor } {}

    waiter(const waiter& rhs): m_waitfactor(rhs.m_waitfactor) {
        slowdown();
    }

    waiter& operator=(const waiter& rhs) {
        m_waitfactor = rhs.m_waitfactor;
        slowdown();
        return *this;
    }
private:
    void slowdown() const {
        for (int i = 0; i < m_waitfactor; ++i) {
            auto func = [=](int d) {
                return i / (d + 1);
            };
            dump = func(dump);
        }
    }
    int m_waitfactor;
};

template <typename QueueType>
long long bench(int wait, int items, int thread_num) {
    QueueType queue;

    auto inserter = [&, items, thread_num]() {
        for (int i = 0; i < items / thread_num; ++i) {
            queue.push(waiter{ wait });
        }
    };
    auto taker = [&, items, thread_num]() {
        waiter w(0);
        for (int i = 0; i < items / thread_num; ++i) {
            queue.pop(w);
        }
    };


    std::vector<std::thread> threads;
    threads.reserve(thread_num * 2);

    auto t1 = std::chrono::high_resolution_clock::now();
    for (int i = 0; i < thread_num; ++i) {
        threads.emplace_back(inserter);
    }
    for (int i = 0; i < thread_num; ++i) {
        threads.emplace_back(taker);
    }

    for (auto& t : threads) {
        t.join();
    }
    auto t2 = std::chrono::high_resolution_clock::now();
    return to_ms(t2 - t1).count();
}

static const int num_iterations = 6;

int main() {
    int items = 320'000;
    std::cout << "Items,NumThreads,WaitFactor,ListQueueMedianMS,NaiveQueueMedianMs\n";
    for (int thread_num : {1, 2, 4, 8, 12, 16, 20, 24}) {
        for (int wait : {0, 1, 10, 100, 250, 500, 750, 1'000}) {
            std::vector<long long> times;
            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(bench<list_queue<waiter>>(wait, items, thread_num));
            }

            // Get median
            std::sort(begin(times), end(times));
            auto lq_median = (times[2] + times[3]) / 2.;

            times.clear();
            for (int iter = 0; iter < num_iterations; ++iter) {
                times.push_back(bench<naive_queue<waiter>>(wait, items, thread_num));
            }

            // Get median
            std::sort(begin(times), end(times));
            auto nq_median = (times[2] + times[3]) / 2.;
            std::cout << items << ',' << thread_num << ',' << wait << ',' << lq_median
                << ',' << nq_median << std::endl;
        }
    }
}
