#include <iostream>
#include <thread>
#include <vector>
#include <string>
#include <atomic>
#include <mutex>
#include <chrono>
#include <random>
#include <algorithm>
#include <numeric>
#include <iterator>


static const int num_size = 256'000'000;

template <typename TimePoint>
std::chrono::milliseconds to_ms(TimePoint tp) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(tp);
}


double generate_double() {
    static std::default_random_engine rng{ std::random_device{}() };
    static std::uniform_real_distribution<> dist(0, 1);
    return dist(rng);
}

std::vector<double> generate_data(int d_size) {
    std::vector<double> ret;
    ret.reserve(d_size);
    std::generate_n(std::back_inserter(ret), d_size, generate_double);
    return ret;
}

static const int num_iterations = 6;

int main(int argc, char** argv) {
    int num_size;
    if (argc == 2) {
        num_size = std::atoi(argv[1]);
    } else {
        num_size = 100'800'000;
    }
    auto data = generate_data(num_size);
    std::cout << "ProblemSize,Threshold,NumThreads,STMedianMs,MTNaiveMedianMs,MTSmartMedianMs\n";
    for (double threshold : {1.0, 1.5, 2.0, 2.5, 3.0}) {
        std::vector<long long> times; times.reserve(num_iterations);
        long long proper_result = 0;
        for (int iter = 0; iter < num_iterations; ++iter) {
            auto ST1 = std::chrono::high_resolution_clock::now();
            proper_result = std::count_if(begin(data), end(data), [=](double d) {
                if (exp(d) > threshold) {
                    return true;
                }
                return false;
            });
            auto ST2 = std::chrono::high_resolution_clock::now();
            times.push_back(to_ms(ST2 - ST1).count());
        }
        auto ST_median = (times[2] + times[3]) / 2.;
        times.clear();
        // LCM: 10080
        for (int num_threads : {2, 4, 8, 12, 16, 20, 24, 28, 32, 36}) {
            std::atomic<int> result1{ 0 }, result2{ 0 };

            auto thread_func1 = [&, threshold](size_t from, size_t to) {
                for (; from < to; ++from) {
                    if (exp(data[from]) > threshold) {
                        result1 += 1;
                    }
                }
            };
            auto thread_func2 = [&, threshold](size_t from, size_t to) {
                int partial_result = 0;
                for (; from < to; ++from) {
                    if (exp(data[from]) > threshold) {
                        partial_result += 1;
                    }
                }
                result2 += partial_result;
            };

            std::vector<std::thread> threads;
            threads.reserve(num_threads);
            auto part_size = num_size / num_threads;

            for (int iter = 0; iter < num_iterations; ++iter) {
                result1 = 0;
                auto MTD1 = std::chrono::high_resolution_clock::now();
                for (int i = 0; i < num_threads; ++i) {
                    threads.emplace_back(thread_func1, i * part_size, (i + 1) * part_size);
                }
                for (auto& t : threads) {
                    t.join();
                }
                auto MTD2 = std::chrono::high_resolution_clock::now();
                times.push_back(to_ms(MTD2 - MTD1).count());
                threads.clear();
            }
            auto MTD_median = (times[2] + times[3]) / 2.;
            times.clear();

            for (int iter = 0; iter < num_iterations; ++iter) {
                result2 = 0;
                auto MTB1 = std::chrono::high_resolution_clock::now();
                for (int i = 0; i < num_threads; ++i) {
                    threads.emplace_back(thread_func2, i * part_size, (i + 1) * part_size);
                }
                for (auto& t : threads) {
                    t.join();
                }
                auto MTB2 = std::chrono::high_resolution_clock::now();
                times.push_back(to_ms(MTB2 - MTB1).count());
                threads.clear();
            }
            auto MTB_median = (times[2] + times[3]) / 2.;
            times.clear();

            if (proper_result != result1 || proper_result != result2) {
                std::cerr << "invalid results!\n";
                std::cerr << num_threads << ' ' << proper_result << ' ' << result1 << ' ' << result2 << '\n';
            } else {
                std::cout << num_size << ',' << threshold << ',' << num_threads << ',' << ST_median << ','
                    << MTD_median << ',' << MTB_median << std::endl;
            }
        }
    }
}
